<?php
    trait CRUD {
        function __construct(){
            if (self::_TABLE === '' OR !is_string(self::_TABLE)) throw new Exception('Crie a constate _TABLE para a classe '.get_called_class());
        }

        /**
         * Retorna uma lista de models que cumprem com as condições passadas em $where
         * 
         * @param array $where ('property' => value) Condições que os registros precisam cumprir para serem incluídos no retorno
         * @param array $campos ('attribute1'[, 'attribute2'[...]]) Colunas que serão pedidas na consulta do banco
         * @param bool $debug Mostrar informações de debug?
         * @param array $inner (tabela => (tabela.id => this.tabela_id)) INNER JOIN com outras tabelas
         * @param array $left (tabela => (tabela.id => this.tabela_id)) LEFT JOIN com outras tabelas
         *
         * @return array
         */
        static function getBy($where, $campos = array('*'), $debug = false, $inner = array(), $left = array(), $groupBy = false, $having = false, $orderBy = false, $limit = false) {
            $sql = 'SELECT '.implode($campos, ',').' FROM '.self::_TABLE;

            if ($inner) {
                $sql .= ' ' . self::makeJoin($inner, 'INNER');
            }

            if ($left) {
                $sql .= ' ' . self::makeJoin($left, 'LEFT');
            }

            $sql .= ' WHERE ';
            self::makeWhere($where, $conditions, $args);

            $sql .= implode($conditions, 'AND');

            $sql .= $groupBy ? " GROUP BY $groupBy" : '';
            $sql .= $having ? " HAVING $having" : '';
            $sql .= $orderBy ? " ORDER BY $orderBy" : '';
            $sql .= $limit ? " LIMIT $limit" : '';

            if ($debug)
                debug_log($sql);

            $Database = new Database();
            $queryResult = $Database->query($sql, $args);
            
            if ($queryResult === false)
                return false;

            $modelClass = 'model'.get_called_class();
            $objsResult = array();
            foreach ($queryResult['result'] as $result) {
                $objsResult[] = new $modelClass($result);
            }
            return $objsResult;
        }

        private static function makeWhere($where, &$conditions, &$args) {
            $conditions = array();
            $args = array();

            if (is_array($where)) {
                foreach ($where as $i => $v) {
                    if (is_array($v)) {
                        $conditions[] = " $i IN (".implode(',', array_fill(0, sizeof($where[$i]), '?')).") ";
                        $args = array_merge($args, $where[$i]);
                    } else if (is_numeric($v)) {
                        $conditions[] = " $i = ? ";
                        $args[] = $where[$i];
                    } else {
                        $conditions[] = " $i LIKE ? ";
                        $args[] = $where[$i];
                    }
                }
            }
        }

        private static function makeJoin($joinArray, $type) {
            $return = '';
            foreach ($joinArray as $tableName => $keys) {
                $selfTable = self::_TABLE;
                $return .= " ${type} JOIN ${tableName} ON ";
                $conds = [];
                foreach ($keys as $k => $v) {
                    $conds[] = "${k} = ${v}";
                }
                
                $return .= implode(' AND ', $conds);
            }
            return $return;
        }

        static function insert($values) {
            $Connection = new Database();

            $sql = 'INSERT INTO ' . self::_TABLE;
            
            $args = array();
            $columns = array();
            foreach ($values as $column => $val) {
                if ($val instanceof Point)
                    continue;
                
                $columns[] = $column;
                if ($val instanceof TimeDate) {
                    $dateVal = $val('Y-m-d H:i:s');
                    if (!$dateVal) 
                        $dateVal = null;
                    $args[] = $dateVal;
                } else if (is_bool($val)) {
                    $args[] = $val ? '1' : '0';
                } else {
                    $args[] = $val;
                }
            }

            $sql .= '(' . implode(',', $columns) . ') VALUES (';
            $sql .= implode(',', array_fill(1, sizeof($args), '?'));
            $sql .= ')';
            $queryResult = $Connection->query($sql, $args);

            if ($queryResult === false)
                return false;

            return $queryResult['lastId'];
        }

        static function update($values, $id) {
            $Connection = new Database();

            $sql = 'UPDATE ' . self::_TABLE . ' SET ';
            
            $args = array();
            $conditions = array();
            foreach ($values as $column => $val) {
                if ($val instanceof Point) {
                    $conditions[] = $column.' = POINT(?, ?)';
                    $args[] = $val->getLat();
                    $args[] = $val->getLon();
                } else if ($val instanceof TimeDate) {
                    $dateVal = $val('Y-m-d H:i:s');
                    if (!$dateVal) 
                        $dateVal = null;
                    $conditions[] = $column.' = ?';
                    $args[] = $dateVal;
                } else {
                    $conditions[] = $column.' = ?';
                    $args[] = $val;
                }
            }

            $sql .= implode(', ', $conditions);
            $sql .= ' WHERE id = ?';
            $args[] = $id;
            
            $queryResult = $Connection->query($sql, $args);

            return (bool)$queryResult;
        }

        static function delete($id) {
            $Connection = new Database();

            $sql = 'UPDATE ' . self::_TABLE . ' SET status = 0 WHERE id = ?';
            $args = array($id);
            $queryResult = $Connection->query($sql, $args);

            return (bool)$queryResult;
        }

        static function reallyDelete($id) {
            $Connection = new Database();

            $sql = 'DELETE FROM ' . self::_TABLE . ' WHERE id = ?';
            $args = array($id);
            $queryResult = $Connection->query($sql, $args);

            return (bool)$queryResult;
        }
    }
?>